/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.commons.lang;

import java.util.Arrays;

/**
 * Helper class for writing implementations of the {@link java.lang.Comparable#compareTo(Object)} method from the
 * {@link java.lang.Comparable} interface. All methods that accept an array as a parameter and methods that accept
 * single Objects as parameters safely handle a <code>null</code> value being passed as arguments. Methods that accept
 * an array also safely handle a <code>null</code> element in the array. The following rules are followed by all
 * compareTo methods:
 *
 * <ul>
 *   <li>Two <code>null</code> values are considered equal.</li>
 *   <li>A <code>null</code> object reference is considered less than a valid object reference. This applies to arrays
 *     as well as regular objects.</li>
 *   <li>An array with a shorter length is considered smaller than one with a longer length, regardless of the contents
 *     of the elements.</li>
 * </ul>
 */
public final class CompareToHelper {

    private CompareToHelper() {
    }

    /**
     * Generates compareTo value for two boolean values. The value <code>true</code> is considered to be greater than
     * the value <code>false</code>.
     *
     * @param   b1  left side of compareTo operation.
     * @param   b2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     */
    public static int compareTo(boolean b1, boolean b2) {
        int result = 0;
        if (b1 != b2) {
            if (b1) {
                result = 1;
            } else {
                result = -1;
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of boolean values.
     *
     * @param   b1  left side of compareTo operation.
     * @param   b2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(boolean[] b1, boolean[] b2) {
        int result = 0;
        if (!Arrays.equals(b1, b2)) {
            if ((b1 == null) && (b2 != null)) {
                result = -1;
            } else if ((b1 != null) && (b2 == null)) {
                result = 1;
            } else if (b1.length != b2.length) {
                result = b1.length - b2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(b1[i], b2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates compareTo value for two long values.
     *
     * @param   l1  left side of compareTo operation.
     * @param   l2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     */
    public static int compareTo(long l1, long l2) {
        int result = 0;
        long diff = l1 - l2;
        if (diff != 0) {
            if (diff > 0) {
                result = 1;
            } else {
                result = -1;
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of long values.
     *
     * @param   l1  left side of compareTo operation.
     * @param   l2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(long[] l1, long[] l2) {
        int result = 0;
        if (!Arrays.equals(l1, l2)) {
            if ((l1 == null) && (l2 != null)) {
                result = -1;
            } else if ((l1 != null) && (l2 == null)) {
                result = 1;
            } else if (l1.length != l2.length) {
                result = l1.length - l2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(l1[i], l2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates compareTo value for two float values.
     *
     * @param   f1  left side of compareTo operation.
     * @param   f2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     */
    public static int compareTo(float f1, float f2) {
        int result = 0;
        float diff = f1 - f2;
        if (diff != 0) {
            if (diff > 0) {
                result = 1;
            } else {
                result = -1;
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of float values.
     *
     * @param   f1  left side of compareTo operation.
     * @param   f2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(float[] f1, float[] f2) {
        int result = 0;
        if (!Arrays.equals(f1, f2)) {
            if ((f1 == null) && (f2 != null)) {
                result = -1;
            } else if ((f1 != null) && (f2 == null)) {
                result = 1;
            } else if (f1.length != f2.length) {
                result = f1.length - f2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(f1[i], f2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates compareTo value for two double values.
     *
     * @param   d1  left side of compareTo operation.
     * @param   d2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     */
    public static int compareTo(double d1, double d2) {
        int result = 0;
        double diff = d1 - d2;
        if (diff != 0) {
            if (diff > 0) {
                result = 1;
            } else {
                result = -1;
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of double values.
     *
     * @param   d1  left side of compareTo operation.
     * @param   d2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(double[] d1, double[] d2) {
        int result = 0;
        if (!Arrays.equals(d1, d2)) {
            if ((d1 == null) && (d2 != null)) {
                result = -1;
            } else if ((d1 != null) && (d2 == null)) {
                result = 1;
            } else if (d1.length != d2.length) {
                result = d1.length - d2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(d1[i], d2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates compareTo value for two int values.
     *
     * @param   i1  left side of compareTo operation.
     * @param   i2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     */
    public static int compareTo(int i1, int i2) {
        return i1 - i2;
    }

    /**
     * Generates a compareTo value for two arrays of int values.
     *
     * @param   i1  left side of compareTo operation.
     * @param   i2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(int[] i1, int[] i2) {
        int result = 0;
        if (!Arrays.equals(i1, i2)) {
            if ((i1 == null) && (i2 != null)) {
                result = -1;
            } else if ((i1 != null) && (i2 == null)) {
                result = 1;
            } else if (i1.length != i2.length) {
                result = i1.length - i2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(i1[i], i2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of byte values.
     *
     * @param   b1  left side of compareTo operation.
     * @param   b2  right side of compareTo operation.
     * @return  the result of comparing each element and combining the results.
     */
    public static int compareTo(byte[] b1, byte[] b2) {
        int result = 0;
        if (!Arrays.equals(b1, b2)) {
            if ((b1 == null) && (b2 != null)) {
                result = -1;
            } else if ((b1 != null) && (b2 == null)) {
                result = 1;
            } else if (b1.length != b2.length) {
                result = b1.length - b2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(b1[i], b2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of char values.
     *
     * @param   c1  left side of compareTo operation.
     * @param   c2  right side of compareTo operation.
     * @return  the result of comparing each element and combining the results.
     */
    public static int compareTo(char[] c1, char[] c2) {
        int result = 0;
        if (!Arrays.equals(c1, c2)) {
            if ((c1 == null) && (c2 != null)) {
                result = -1;
            } else if ((c1 != null) && (c2 == null)) {
                result = 1;
            } else if (c1.length != c2.length) {
                result = c1.length - c2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(c1[i], c2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two arrays of short values.
     *
     * @param   s1  left side of compareTo operation.
     * @param   s2  right side of compareTo operation.
     * @return  the result of comparing each element and combinging the results.
     */
    public static int compareTo(short[] s1, short[] s2) {
        int result = 0;
        if (!Arrays.equals(s1, s2)) {
            if ((s1 == null) && (s2 != null)) {
                result = -1;
            } else if ((s1 != null) && (s2 == null)) {
                result = 1;
            } else if (s1.length != s2.length) {
                result = s1.length - s2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(s1[i], s2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates compareTo value for two Comparable values.
     *
     * @param   c1   left side of compareTo operation.
     * @param   c2   right side of compareTo operation.
     * @return  the resulting compareTo value.
     * @param <T> a T object.
     */
    public static <T extends Comparable<T>> int compareTo(T c1, T c2) {
        int result = 0;
        if ((c1 != null) && (c2 != null)) {
            result = c1.compareTo(c2);
        } else if ((c1 == null) && (c2 != null)) {
            result = -1;
        } else if ((c1 != null) && (c2 == null)) {
            result = 1;
        }
        return result;
    }

    /**
     * Generates a compareTo value for an array of Comparable objects.
     *
     * @param   c1   left side of compareTo operation.
     * @param   c2   right side of compareTo operation.
     * @return  the result of comparing each element and combining the results.
     * @param <T> a T object.
     */
    public static <T extends Comparable<T>> int compareTo(T[] c1, T[] c2) {
        int result = 0;
        if (!Arrays.equals(c1, c2)) {
            if ((c1 == null) && (c2 != null)) {
                result = -1;
            } else if ((c1 != null) && (c2 == null)) {
                result = 1;
            } else if (c1.length != c2.length) {
                result = c1.length - c2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(c1[i], c2[i]);
                    i++;
                }
            }
        }
        return result;
    }

    /**
     * Generates a compareTo value for two Objects. This method is only fully reliable when one of the following
     * conditions is true:
     *
     * <ul>
     *   <li>both arguments are null</li>
     *   <li>one argument is null</li>
     *   <li>the Class of the arguments overrides the default implementation of the hashCode method in such a way that
     *     naturally greater objects have a larger hashCode value than smaller objects and vice versa</li>
     * </ul>
     *
     * <p>If none of these conditions are met, then this method relies on the equals and hashCode methods of the passed
     * in Classes and so the return value is not guaranteed to adhere to the general contract of the Comparable
     * interface. The {@link #compareTo(Comparable, Comparable)} method should be used instead where possible.</p>
     *
     * @param   o1  left side of compareTo operation.
     * @param   o2  right side of compareTo operation.
     * @return  the resulting compareTo value.
     * @throws  java.lang.ClassCastException  if both arguments are not <code>null</code> and the following condition is not true
     *                              <code>o1.getClass().isAssignableFrom(o2.getClass())</code>
     */
    public static int compareTo(Object o1, Object o2) {
        int result = 0;
        if ((o1 != null) && (o2 != null)) {
            if (!o1.getClass().isAssignableFrom(o2.getClass())) {
                throw new ClassCastException(o1.getClass() + " is not assignable from " + o2.getClass());
            } else if (!o1.equals(o2)) {
                result = o1.hashCode() - o2.hashCode();
            }
        } else if ((o1 == null) && (o2 != null)) {
            result = -1;
        } else if ((o1 != null) && (o2 == null)) {
            result = 1;
        }
        return result;
    }

    /**
     * Generates a compareTo value for an array of Objects.
     *
     * @param   o1  left side of compareTo operation.
     * @param   o2  right side of compareTo operation.
     * @return  the result of comparing each element and combining the results.
     * @see     #compareTo(Object, Object)
     */
    public static int compareTo(Object[] o1, Object[] o2) {
        int result = 0;
        if (!Arrays.equals(o1, o2)) {
            if ((o1 == null) && (o2 != null)) {
                result = -1;
            } else if ((o1 != null) && (o2 == null)) {
                result = 1;
            } else if (o1.length != o2.length) {
                result = o1.length - o2.length;
            } else {
                // at this point we know that the arrays are the same length but at least one
                // element must be different
                int i = 0;
                while (result == 0) {
                    result = compareTo(o1[i], o2[i]);
                    i++;
                }
            }
        }
        return result;
    }
}
