/*
 * Copyright 2010-2011 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.commons.lang;

import java.io.File;

import java.lang.reflect.Array;

import java.net.URL;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;

import com.squeakysand.commons.beans.BeanHelper;
import com.squeakysand.commons.beans.BeanHelperException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Provides utility methods to help override the {@link Object#toString()} method and/or create more meaningful
 * logging output.
 */
public final class ToStringHelper {

    private static final Logger LOG = LoggerFactory.getLogger(ToStringHelper.class);
    private static final String NULL = "null";
    private static final List<StringRenderer> RENDERERS = new ArrayList<StringRenderer>();

    static {
        RENDERERS.add(new StringStringRenderer());
        RENDERERS.add(new FloatStringRenderer());
        RENDERERS.add(new LongStringRenderer());
        RENDERERS.add(new ArrayStringRenderer());
        RENDERERS.add(new MapStringRenderer());
        RENDERERS.add(new CollectionStringRenderer());
        RENDERERS.add(new UrlStringRenderer());
        RENDERERS.add(new FileStringRenderer());
        RENDERERS.add(new ValueOfStringRenderer());
        RENDERERS.add(new JavaBeanStringRenderer());
    }

    private ToStringHelper() {
    }

    private static StringRenderer findRenderer(Class<?> klass) {
        // check to see if there is a converter directly registered for this class
        StringRenderer result = null;
        for (StringRenderer renderer : RENDERERS) {
            if (renderer.isAbleToRender(klass)) {
                result = renderer;
                break;
            }
        }
        LOG.debug("for {}, renderer is {}", klass.getName(), result.getClass().getName());
        return result;
    }

    /**
     * Generates a String representation of the passed in Object. This method will handle <code>null</code> values,
     * objects, arrays of primitives and arrays of objects.
     *
     * @param   o  the target object, can be <code>null</code> or an array.
     * @return  a String representation of the target object.
     */
    public static String toString(Object o) {
        String result = null;
        if (o == null) {
            result = NULL;
        } else {
            result = findRenderer(o.getClass()).render(o);
        }
        return result;
    }

    /**
     * Generates a String representation of the state of the JavaBean passed in. Includes only the read-write properties
     * (equivalent to calling <code>getPropertyString(o, false)</code>).
     *
     * @param   bean  the JavaBean.
     * @return  a formatted String representing the values of the JavaBean properties of the bean.
     */
    public static String getPropertiesString(Object bean) {
        return getPropertiesString(bean, false);
    }

    /**
     * Generates a String representation of the state of the JavaBean passed in.
     *
     * @param   bean                          the JavaBean.
     * @param   includeReadOnlyProperties  indicates if read-only properties should be included.
     * @return  a formatted String representing the values of the JavaBean properties of the bean.
     */
    public static String getPropertiesString(Object bean, boolean includeReadOnlyProperties) {
        String result = null;
        if (bean != null) {
            try {
                Map<String, Object> properties = BeanHelper.getPropertyValues(bean, includeReadOnlyProperties);
                result = toString(properties);
            } catch (BeanHelperException e) {
                LOG.error(e.getMessage(), e);
            }
        }
        return result;
    }

    private static interface StringRenderer {

        String render(Object target);

        boolean isAbleToRender(Class<?> klass);
    }

    private static class ArrayStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            StringBuilder sb = new StringBuilder();
            int length = Array.getLength(target);
            sb.append('[');
            for (int i = 0; i < length; i++) {
                Object element = Array.get(target, i);
                sb.append(ToStringHelper.toString(element));
                if (i != (length - 1)) {
                    sb.append(", ");
                }
            }
            sb.append(']');
            return sb.toString();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return klass.isArray();
        }
    }

    private static class CollectionStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            StringBuilder sb = new StringBuilder("(");
            Collection<?> collection = (Collection<?>) target;
            for (Object o : collection) {
                sb.append(ToStringHelper.toString(o));
                sb.append(", ");
            }
            // chop off the last comma separator
            if (sb.length() > 1) {
                sb.setLength(sb.length() - 2);
            }
            sb.append(")");
            return sb.toString();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return Collection.class.isAssignableFrom(klass);
        }
    }

    private static class FileStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return ((File) target).getPath();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return File.class.isAssignableFrom(klass);
        }
    }

    private static class FloatStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return Float.toString((Float) target) + 'F';
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return Float.class.isAssignableFrom(klass) || float.class.isAssignableFrom(klass);
        }
    }

    private static class JavaBeanStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            StringBuilder sb = new StringBuilder();
            String name = target.getClass().getName();
            int dotIndex = name.lastIndexOf('.');
            if (dotIndex != -1) {
                name = name.substring(dotIndex + 1);
            }
            sb.append(name);
            sb.append(getPropertiesString(target));
            return sb.toString();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return true;
        }
    }

    private static class LongStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return Long.toString((Long) target) + 'L';
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return Long.class.isAssignableFrom(klass) || long.class.isAssignableFrom(klass);
        }
    }

    private static class MapStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            Map<?, ?> map = (Map<?, ?>) target;
            StringBuilder sb = new StringBuilder("{");
            for (Object key : map.keySet()) {
//              sb.append("{");
                sb.append(ToStringHelper.toString(key));
                sb.append("=");
                sb.append(ToStringHelper.toString(map.get(key)));
//              sb.append("}");
                sb.append(", ");
            }
            // chop off the last comma separator
            if (sb.length() > 1) {
                sb.setLength(sb.length() - 2);
            }
            sb.append("}");
            return sb.toString();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return Map.class.isAssignableFrom(klass);
        }
    }

    private static class StringStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return '\"' + (String) target + '\"';
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return String.class.isAssignableFrom(klass);
        }
    }

    private static class UrlStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return ((URL) target).toExternalForm();
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return URL.class.isAssignableFrom(klass);
        }
    }

    private static class ValueOfStringRenderer implements StringRenderer {

        @Override
        public String render(Object target) {
            return String.valueOf(target);
        }

        @Override
        public boolean isAbleToRender(Class<?> klass) {
            return klass.isPrimitive() || klass.isEnum() || klass.isAnnotation() || Boolean.class.isAssignableFrom(klass)
                   || Character.class.isAssignableFrom(klass) || Class.class.isAssignableFrom(klass) || Date.class.isAssignableFrom(klass)
                   || Number.class.isAssignableFrom(klass);
        }
    }
}
